// ------------------------------------------------------------------------------------------------------------------------------------------------------------
//
//	File: encoder.c
//
//	Quadrature Encoder for Orb Motor Control Unit
//
//	Version 14.3
//	7-May-2007
// ------------------------------------------------------------------------------------------------------------------------------------------------------------
 
#include <avr/io.h>
#include "encoder.h"   

/* static variables for encoder routines */

static short Quad_Count;					// see below, could be SIGNED char
static uint8_t prevPin1, prevPin2;

#define MAX_SAMPLES	10
static short prevSample[MAX_SAMPLES];		// depending on rotation speed, these
											// could be changed to 8Bit chars
											
// ------------------------------------------------------------------------------------------------------------------------------------------------------------
// Setup inputs for polling Quad Encoder Pins
// Pins PortD 2:3 are input pins.

void Encoder_Init(void)
{
	uint8_t n;

	for (n=0;n<MAX_SAMPLES;n++)
		prevSample[n] = 0;

	Quad_Count = 0;
	prevPin1 = 0;
	prevPin2 = 0;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------------
// Encoder_poll_pins()
// Called from Main Loop -
// Poll the encoder pins 160,000 times per second.  (at 8.0 Mhz)
// Check for pin state changes - Inc / Dec Quad_Count
// Decode Quad encoder pins based on Standard Quad Encoder truth table:
//
// prev  crnt	Direction
// -----------------------------------
//	00 -> 01	-1
//	01 -> 11	-1
//	10 -> 00	-1
//	11 -> 10	-1
//
//	00 -> 10	+1
//	01 -> 00	+1
//	10 -> 11	+1
//	11 -> 01	+1

void Encoder_poll_pins(void)
{
	uint8_t crntPin1, crntPin2, thePins;
	
	thePins = PIND;		// read port D pins 2 & 3
	crntPin1 = (thePins & 0x04) >> 2;	// 0b 0100
	crntPin2 = (thePins & 0x08) >> 3;	// 0b 1000
	
	if ((crntPin1 != prevPin1) || (crntPin2 != prevPin2))	// something changed
		{	// check for both pins changed - error
		if (!((crntPin1 != prevPin1) && (crntPin2 != prevPin2)))	
			{
			if (crntPin1 == prevPin2)
				Quad_Count++;
			else
				Quad_Count--;
			}
		prevPin1 = crntPin1;
		prevPin2 = crntPin2;
		}
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------------
// Called from Main Loop -
// Sample the encoder 10 times a second.  
// during testing, I'm using a wheel encoder that's only spinning at 60 rpm
// this only gives us a few clicks per 100ms sample -- too few.
// because I'm using a very slow motor, I'm agregating the samples over 1 second.

void Encoder_sample_speed(void)
{
	uint8_t n;
	
	for (n=0;n<(MAX_SAMPLES-1);n++)
		prevSample[n] = prevSample[n+1];
		
	prevSample[MAX_SAMPLES-1] = Quad_Count;
	Quad_Count = 0;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------------
// return the sum of the prev 5 or 10 samples.
// this increases our resolution, but introduces a time lag in the feedback.

unsigned short Encoder_read_speed(void)
{
	unsigned char n;
	unsigned short theCount = 0;
	
	for (n=0;n<(MAX_SAMPLES-1);n++)
		theCount += prevSample[n];

	return theCount;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------------
// End of File
